/*
 * Copyright (C) 2008 Andre Noll <maan@tuebingen.mpg.de>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file user.c \brief User and user ID handling. */

#include "adu.h"
#include <dirent.h> /* readdir() */
#include <sys/types.h>
#include <pwd.h>
#include "user.h"
#include "fd.h"
#include "string.h"
#include "error.h"

/**
 * Describes one range of admissible user IDs.
 *
 * adu converts the admissible user ids given at the command line
 * into an array of such structs.
 */
struct uid_range {
	/** Lowest admissible user ID. */
	uint32_t low;
	/** Greatest admissible user ID. */
	uint32_t high;
};

/** Iterate over all uid ranges. */
#define FOR_EACH_UID_RANGE(ur, urs) for (ur = urs; ur->low <= ur->high; ur++)

/** Flags for the user hash table. */
enum uid_info_flags {
	/** Whether this slot of the hash table is used. */
	UI_FL_SLOT_USED = 1,
	/** Whether this uid should be taken into account. */
	UI_FL_ADMISSIBLE = 2,
};

/*
 * Contains info for each user that owns at least one regular file.
 *
 * Even users that are not taken into account because of the --uid
 * option occupy a slot in this hash table. This allows to find out
 * quicky whether a uid is admissible. And yes, this has to be fast.
 */
static struct user_info *uid_hash_table;

/** This is always a power of two. It is set in create_hash_table(). */
static uint32_t uid_hash_table_size;

/** The number of used slots in the hash table. */
static uint32_t num_uids;

/*
 * The columns of the per-user tables.
 *
 * Adu tracks disk usage on a per-user basis. For each user, a user table is
 * being created. The rows of the user table have three columns: The directory
 * number that may be resolved to the path using the directory table, the
 * number of bytes and the number of files in that directory owned by the given
 * user.
 */
static struct osl_column_description user_table_cols[] = {
	[UT_DIR_NUM] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE | OSL_UNIQUE,
		.name = "dir_num",
		.compare_function = uint64_compare,
		.data_size = sizeof(uint64_t)
	},
	[UT_BYTES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_bytes",
		.data_size = sizeof(uint64_t)
	},
	[UT_FILES] = {
		.storage_type = OSL_MAPPED_STORAGE,
		.storage_flags = OSL_RBTREE | OSL_FIXED_SIZE,
		.compare_function = size_compare,
		.name = "num_files",
		.data_size = sizeof(uint64_t)
	},
};

static int check_uid_arg(const char *arg, uint32_t *uid)
{
	const uint32_t max = ~0U;
	/*
	 * we need an 64-bit int for string -> uid conversion because strtoll()
	 * returns a signed value.
	 */
	int64_t val;
	int ret = atoi64(arg, &val);

	if (ret < 0)
		return ret;
	if (val < 0 || val > max)
		return -ERRNO_TO_ERROR(EINVAL);
	*uid = val;
	return 1;
}

static int parse_uid_range(const char *orig_arg, struct uid_range *ur)
{
	int ret;
	char *arg = adu_strdup(orig_arg), *p = strchr(arg, '-');

	if (!p || p == arg) { /* -42 or 42 */
		ret = check_uid_arg(p? p + 1 : arg, &ur->high);
		if (ret < 0)
			goto out;
		ur->low = p? 0 : ur->high;
		ret = 1;
		goto out;
	}
	/* 42- or 42-4711 */
	*p = '\0';
	p++;
	ret = check_uid_arg(arg, &ur->low);
	if (ret < 0)
		goto out;
	ur->high = ~0U;
	if (*p) { /* 42-4711 */
		ret = check_uid_arg(p, &ur->high);
		if (ret < 0)
			goto out;
	}
	if (ur->low > ur->high)
		ret = -ERRNO_TO_ERROR(EINVAL);
out:
	if (ret < 0)
		ERROR_LOG("bad uid option: %s\n", orig_arg);
	else
		INFO_LOG("admissible uid range: %u - %u\n", ur->low,
			ur->high);
	free(arg);
	return ret;
}

/**
 * Convert the --uid argument to an array of uid ranges.
 *
 * \param orig_arg The argument to the --uid option.
 * \param ur Result pointer.
 *
 * Returns Negative on errors. On success, the number of uid ranges
 * is returned.
 */
int parse_uid_arg(const char *orig_arg, struct uid_range **ur)
{
	char *arg, **argv;
	unsigned n;
	int i, ret = 1;

	if (!orig_arg)
		return 0;
	arg = adu_strdup(orig_arg);
	n = split_args(arg, &argv, ",");
	if (!n)
		return -E_SYNTAX;
	*ur = adu_malloc((n + 1) * sizeof(struct uid_range));
	for (i = 0; i < n; i++) {
		ret = parse_uid_range(argv[i], *ur + i);
		if (ret < 0)
			break;
	}
	free(argv);
	free(arg);
	if (ret < 0) {
		free(*ur);
		*ur = NULL;
		return ret;
	}
	/* an empty range indicates the end of the list */
	(*ur)[n].low = 1;
	(*ur)[n].high = 0;
	return n;
}

static int uid_is_admissible(uint32_t uid, struct uid_range *urs)
{
	struct uid_range *ur;
	int ret = 1;

	if (!urs) /* empty array means all uids are allowed */
		return 1;
	FOR_EACH_UID_RANGE(ur, urs)
		if (ur->low <= uid && ur->high >= uid)
			goto out;
	ret = 0;
out:
	DEBUG_LOG("uid %u is %sadmissible\n", (unsigned)uid,
		ret? "" : "not ");
	return ret;
}

/**
 * Add each given user to the array of admissible users.
 *
 * \param users Array of user names to add.
 * \param num_users Length of \a users.
 * \param admissible_uids The users which are already admissible.
 * \param num_uid_ranges The number of intervals of \a admissible_uids.
 *
 * For each given user, the function checks whether that user is already
 * admissible, i.e. its uid is contained in one of the ranges given by \a
 * admissible_uids. If it is, the function ignores that user. Otherwise, a new
 * length-one range consisting of that uid only is appended to \a
 * admissible_uids.
 *
 * \return Negative on errors, the new number of uid ranges on success.
 */
int append_users(char **users, int num_users,
		struct uid_range **admissible_uids, int num_uid_ranges)
{
	int i;
	struct uid_range *au = *admissible_uids;

	for (i = 0; i < num_users; i++) {
		char *u = users[i];
		struct uid_range *ur;
		struct passwd *pw = getpwnam(u);

		if (!pw) {
			ERROR_LOG("user %s not found\n", u);
			return -ERRNO_TO_ERROR(EINVAL);
		}
		if (au && uid_is_admissible(pw->pw_uid, au))
			continue; /* nothing to do */
		/* add a range consisting of this uid only */
		num_uid_ranges++;
		au = adu_realloc(au, (num_uid_ranges + 1) *
			sizeof(struct uid_range));
		*admissible_uids = au;
		ur = au + num_uid_ranges - 1; /* the new uid range */
		ur->low = ur->high = pw->pw_uid;
		/* terminate the list */
		ur++;
		ur->low = 1;
		ur->high = 0;
	}
	return num_uid_ranges;
}

static inline int ui_used(struct user_info *ui)
{
	return ui->flags & UI_FL_SLOT_USED;
}

static inline int ui_admissible(struct user_info *ui)
{
	return ui->flags & UI_FL_ADMISSIBLE;
}

static int open_user_table(struct user_info *ui, int create)
{
	int ret;
	struct passwd *pw;

	ui->desc = adu_malloc(sizeof(*ui->desc));
	ui->desc->num_columns = NUM_UT_COLUMNS;
	ui->desc->flags = 0;
	ui->desc->column_descriptions = user_table_cols;
	ui->desc->dir = adu_strdup(database_dir);
	ui->desc->name = make_message("%u", (unsigned)ui->uid);
	pw = getpwuid(ui->uid);
	if (pw && pw->pw_name)
		ui->pw_name = adu_strdup(pw->pw_name);

	INFO_LOG("opening table for uid %u\n", (unsigned)ui->uid);
	if (create) {
		ret = osl(osl_create_table(ui->desc));
		if (ret < 0)
			goto err;
		num_uids++;
	}
	ret = osl(osl_open_table(ui->desc, &ui->table));
	if (ret < 0)
		goto err;
	return 1;
err:
	free((char *)ui->desc->name);
	free((char *)ui->desc->dir);
	free(ui->pw_name);
	free(ui->desc);
	ui->desc->name = NULL;
	ui->desc->dir = NULL;
	ui->desc = NULL;
	ui->table = NULL;
	ui->flags = 0;
	return ret;
}

/** Iterate over each user in the uid hash table. */
#define FOR_EACH_USER(ui) for (ui = uid_hash_table; ui < \
	uid_hash_table + uid_hash_table_size; ui++)


/**
 * Execute the given function for each admissible user.
 *
 * \param func The function to execute.
 * \param data Arbitrary pointer.
 *
 * This function calls \a func for each admissible user in the uid hash table.
 * The \a data pointer is passed as the second argument to \a func.
 *
 * \return As soon as \a func returns a negative value, the loop is terminated
 * and that negative value is returned. Otherwise, the function returns 1.
 */
int for_each_admissible_user(int (*func)(struct user_info *, void *),
		void *data)
{
	int i;

	assert(uid_hash_table);
	for (i = 0; i < uid_hash_table_size; i++) {
		int ret;
		struct user_info *ui = uid_hash_table + i;

		if (!ui_used(ui) || !ui_admissible(ui))
			continue;
		ret = func(ui, data);
		if (ret < 0)
			return ret;
	}
	return 1;
}

/** Prime number used for calculating the slot for an uid. */
#define PRIME1 0xb11924e1
/** Prime number used for probe all slots. */
#define PRIME2 0x01000193

/**
 * Create a hash table large enough of given size.
 *
 * \param bits Sets the maximal number of hash table entries to ^bits.
 */
void create_hash_table(unsigned bits)
{
	uid_hash_table_size = 1 << bits;
	uid_hash_table = adu_calloc(uid_hash_table_size *
		sizeof(struct user_info));
}

/**
 * Close all open user tables and destroy the uid hash table.
 *
 * For each used slot in the uid hash table, close the osl user table if it is
 * open.  Finally, free the uid hash table.
 */
void close_user_tables(void)
{
	struct user_info *ui;

	FOR_EACH_USER(ui) {
		int ret;

		if (!ui_used(ui))
			continue;
		if (!ui->table)
			continue;
		INFO_LOG("closing user table for uid %u\n", (unsigned)ui->uid);
		ret = osl(osl_close_table(ui->table, OSL_MARK_CLEAN));
		if (ret < 0)
			ERROR_LOG("failed to close user table %u: %s\n",
				(unsigned)ui->uid, adu_strerror(-ret));
		free((char *)ui->desc->name);
		ui->desc->name = NULL;
		free((char *)ui->desc->dir);
		ui->desc->dir = NULL;
		free(ui->pw_name);
		ui->pw_name = NULL;
		free(ui->desc);
		ui->desc = NULL;
		ui->table = NULL;
		ui->flags = 0;
	}
	free(uid_hash_table);
	uid_hash_table = NULL;
}

/*
 * We use a hash table of size s=2^uid_hash_bits to map the uids into the
 * interval [0..s-1]. Hash collisions are treated by open addressing, i.e.
 * unused slots in the table are used to store different uids that hash to the
 * same slot.
 *
 * If a hash collision occurs, different slots are successively probed in order
 * to find an unused slot for the new uid. Probing is implemented via a second
 * hash function that maps the uid to h=(uid * PRIME2) | 1, which is always an
 * odd number.
 *
 * An odd number is sufficient to make sure each entry of the hash table gets
 * probed for probe_num between 0 and s-1 because s is a power of two, hence
 * the second hash value has never a common divisor with the hash table size.
 * IOW: h is invertible in the ring [0..s-1].
 */
static uint32_t double_hash(uint32_t uid, uint32_t probe_num)
{
	return (uid * PRIME1 + ((uid * PRIME2) | 1) * probe_num)
		% uid_hash_table_size;
}

static struct user_info *lookup_uid(uint32_t uid)
{
	uint32_t p;

	for (p = 0; p < uid_hash_table_size; p++) {
		struct user_info *ui = uid_hash_table + double_hash(uid, p);
		if (!ui_used(ui))
			return ui;
		if (ui->uid == uid)
			return ui;
	}
	return NULL;
}

/**
 * Create and open a osl table for the given uid.
 *
 * \param uid The user ID.
 * \param ui_ptr Result pointer
 *
 * Find out whether \a uid already exists in the uid hash table.  If yes, just
 * return the user info struct via \a ui_ptr. Otherwise, insert \a uid into the
 * uid hash table, create and open the osl user table and also return the user
 * info struct via \a ui_ptr.
 *
 * \return Standard.
 */
int create_user_table(uint32_t uid, struct user_info **ui_ptr)
{
	struct user_info *ui = lookup_uid(uid);

	if (!ui)
		return -E_HASH_TABLE_OVERFLOW;
	*ui_ptr = ui;
	if (ui_used(ui))
		return 1;
	ui->uid = uid;
	ui->flags |= UI_FL_SLOT_USED;
	return open_user_table(ui, 1);
}

static char *get_uid_list_name(void)
{
	return make_message("%s/uid_list", database_dir);
}
/**
 * Open the osl tables for all admissible uids.
 *
 * \param admissible_uids Determines which uids are considered admissible.
 *
 * Each slot in the hash table contains, among other information, a bit which
 * specifies whether the uid of the slot is admissible in the current context.
 *
 * This function iterates over all entries in the hash table and checks for
 * each used slot whether the corresponding uid is admissible with respect to
 * \a admissible_uids. If so, it sets the admissible bit for this slot and
 * opens the osl table of the uid.
 *
 * \return Stamdard.
 */
int open_admissible_user_tables(struct uid_range *admissible_uids)
{
	struct user_info *ui;

	assert(uid_hash_table);
	DEBUG_LOG("size: %d\n", uid_hash_table_size);
	FOR_EACH_USER(ui) {
		int ret;

		if (!ui_used(ui))
			continue;
		if (!uid_is_admissible(ui->uid, admissible_uids)) {
			DEBUG_LOG("uid %u is not admissible\n", ui->uid);
			ui->flags &= ~UI_FL_ADMISSIBLE;
			continue;
		}
		ui->flags |= UI_FL_ADMISSIBLE;
		if (ui->table)
			continue;
		ret = open_user_table(ui, 0);
		if (ret < 0)
			return ret;
	}
	return 1;
}

/**
 * Read the file of all possible uids.
 *
 * This is called from select/interactive mode. First a large hash table, large
 * enough to store all uids contained in the uid file is created. Next, the
 * uids are read from the uid file which was created during the creation of the
 * database and each uid is inserted into the hash table.
 *
 * \sa write_uid_file().
 *
 * \return Standard.
 */
int read_uid_file(void)
{
	size_t size;
	uint32_t n;
	char *filename = get_uid_list_name(), *map;
	int ret = mmap_full_file(filename, O_RDONLY, (void **)&map, &size, NULL);
	unsigned bits;

	if (ret < 0) {
		ERROR_LOG("failed to map %s\n", filename);
		free(filename);
		return ret;
	}
	num_uids = size / 4;
	INFO_LOG("found %u uids in %s\n", (unsigned)num_uids, filename);
	free(filename);
	/*
	 * Compute number of hash table bits. The hash table size must be a
	 * power of two and larger than the number of uids.
	 */
	bits = 2;
	while (1 << bits < num_uids)
		bits++;
	create_hash_table(bits);
	for (n = 0; n < num_uids; n++) {
		uint32_t uid = read_u32(map + n * sizeof(uid));
		struct user_info *ui = lookup_uid(uid);
		assert(ui);
		if (ui_used(ui)) { /* impossible */
			ERROR_LOG("duplicate user id!?\n");
			ret = -EFAULT;
			goto out;
		}
		ui->uid = uid;
		ui->flags |= UI_FL_SLOT_USED;
	}
	ret = 1;
out:
	adu_munmap(map, size);
	return ret;
}

/**
 * Write the list of uids to permanent storage.
 *
 * This is called from create mode after the dir table and all uer tables have
 * been created. The file simply contains the list of all uids that own at
 * least one regular file in the base directory and hence an osl table for this
 * uid exists.
 *
 * \sa read_uid_file().
 *
 * \return Standard.
 */
int write_uid_file(void)
{
	char *buf, *p, *filename;
	size_t size = num_uids * sizeof(uint32_t);
	int ret;
	struct user_info *ui;

	if (!num_uids)
		return 0;
	buf = p = adu_malloc(size);
	FOR_EACH_USER(ui) {
		if (!ui_used(ui))
			continue;
		write_u32(p, ui->uid);
		p += sizeof(uint32_t);
	}
	filename = get_uid_list_name();
	ret = adu_write_file(filename, buf, size);
	free(filename);
	free(buf);
	return ret;
}
